/*
 * Clock and power management. This exposes the combined functionality
 * provided by the jz4780 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-jz4780.h"



// Register locations.

enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR

  Divider_bch         = 0x0ac,  // BCHCDR
  Divider_cim         = 0x07c,  // CIMCDR
  Divider_ddr         = 0x02c,  // DDRCDR
  Divider_gpu         = 0x088,  // GPUCDR
  Divider_hdmi        = 0x08c,  // HDMICDR
  Divider_i2s0        = 0x060,  // I2SCDR
  Divider_i2s1        = 0x0a0,  // I2S1CDR
  Divider_lcd0        = 0x054,  // LP0CDR
  Divider_lcd1        = 0x064,  // LP1CDR
  Divider_msc0        = 0x068,  // MSC0CDR
  Divider_msc1        = 0x0a4,  // MSC1CDR
  Divider_msc2        = 0x0a8,  // MSC2CDR
  Divider_pcm         = 0x084,  // PCMCDR
  Divider_ssi         = 0x074,  // SSICDR
  Divider_uhc         = 0x06c,  // UHCCDR
  Divider_vpu         = 0x030,  // VPUCDR

  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_scratch         = 0x034,  // CPSPR
  Cpm_scratch_prot    = 0x038,  // CPSPPR

  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1

  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_control_V       = 0x01c,  // CPVPCR
};



// Register field definitions.

static Field Clock_source_main         (Clock_control, 3, 30),       // SEL_SRC (output to SCLK_A)
             Clock_source_cpu          (Clock_control, 3, 28),       // SEL_CPLL (output to CCLK)
             Clock_source_hclock0      (Clock_control, 3, 26),       // SEL_H0PLL (output to AHB0)
             Clock_source_hclock2      (Clock_control, 3, 24),       // SEL_H2PLL (output to AHB2)
             Clock_source_bch          (Divider_bch, 3, 30),         // BPCS
             Clock_source_cim          (Divider_cim, 1, 31),         // CIMPCS
             Clock_source_ddr          (Divider_ddr, 3, 30),         // DCS
             Clock_source_gpu          (Divider_gpu, 3, 30),         // GPCS
             Clock_source_hdmi         (Divider_hdmi, 3, 30),        // HPCS
             Clock_source_i2s0         (Divider_i2s0, 3, 30),        // I2CS, I2PCS
             Clock_source_i2s1         (Divider_i2s1, 3, 30),        // I2CS, I2PCS
             Clock_source_lcd0         (Divider_lcd0, 3, 30),        // LPCS
             Clock_source_lcd1         (Divider_lcd1, 3, 30),        // LPCS
             Clock_source_msc          (Divider_msc0, 3, 30),        // MPCS
             Clock_source_pcm          (Divider_pcm, 7, 29),         // PCMS, PCMPCS
             Clock_source_ssi          (Divider_ssi, 1, 30),         // SPCS
             Clock_source_uhc          (Divider_uhc, 3, 30),         // UHCS
             Clock_source_vpu          (Divider_vpu, 3, 30),         // VCS

             Clock_busy_cpu            (Clock_status, 1, 0),
             Clock_busy_hclock0        (Clock_status, 1, 1),
             Clock_busy_hclock2        (Clock_status, 1, 2),
             Clock_busy_bch            (Divider_bch, 1, 28),
             Clock_busy_cim            (Divider_cim, 1, 29),
             Clock_busy_ddr            (Divider_ddr, 1, 28),
             Clock_busy_gpu            (Divider_gpu, 1, 28),
             Clock_busy_hdmi           (Divider_hdmi, 1, 28),
             Clock_busy_i2s0           (Divider_i2s0, 1, 28),
             Clock_busy_i2s1           (Divider_i2s1, 1, 28),
             Clock_busy_lcd0           (Divider_lcd0, 1, 27),
             Clock_busy_lcd1           (Divider_lcd1, 1, 27),
             Clock_busy_msc0           (Divider_msc0, 1, 28),
             Clock_busy_msc1           (Divider_msc1, 1, 28),
             Clock_busy_msc2           (Divider_msc2, 1, 28),
             Clock_busy_pcm            (Divider_pcm, 1, 27),
             Clock_busy_ssi            (Divider_ssi, 1, 28),
             Clock_busy_uhc            (Divider_uhc, 1, 28),
             Clock_busy_vpu            (Divider_vpu, 1, 28),

             Clock_change_enable_cpu   (Clock_control, 1, 22),
             Clock_change_enable_ahb0  (Clock_control, 1, 21),
             Clock_change_enable_ahb2  (Clock_control, 1, 20),
             Clock_change_enable_bch   (Divider_bch, 1, 29),
             Clock_change_enable_cim   (Divider_cim, 1, 30),
             Clock_change_enable_ddr   (Divider_ddr, 1, 29),
             Clock_change_enable_gpu   (Divider_gpu, 1, 29),
             Clock_change_enable_hdmi  (Divider_hdmi, 1, 29),
             Clock_change_enable_i2s0  (Divider_i2s0, 1, 29),
             Clock_change_enable_i2s1  (Divider_i2s1, 1, 29),
             Clock_change_enable_lcd0  (Divider_lcd0, 1, 28),
             Clock_change_enable_lcd1  (Divider_lcd1, 1, 28),
             Clock_change_enable_msc0  (Divider_msc0, 1, 29),
             Clock_change_enable_msc1  (Divider_msc1, 1, 29),
             Clock_change_enable_msc2  (Divider_msc2, 1, 29),
             Clock_change_enable_pcm   (Divider_pcm, 1, 28),
             Clock_change_enable_ssi   (Divider_ssi, 1, 29),
             Clock_change_enable_uhc   (Divider_uhc, 1, 29),
             Clock_change_enable_vpu   (Divider_vpu, 1, 29),

             Clock_divider_cpu         (Clock_control, 0x0f, 0),     // CDIV
             Clock_divider_hclock0     (Clock_control, 0x0f, 8),     // H0DIV (fast AHB peripherals)
             Clock_divider_hclock2     (Clock_control, 0x0f, 12),    // H2DIV (fast AHB peripherals)
             Clock_divider_l2cache     (Clock_control, 0x0f, 4),     // L2CDIV
             Clock_divider_pclock      (Clock_control, 0x0f, 16),    // PDIV (slow APB peripherals)
             Clock_divider_bch         (Divider_bch, 0x0f, 0),       // BCHCDR
             Clock_divider_cim         (Divider_cim, 0xff, 0),       // CIMCDR
             Clock_divider_ddr         (Divider_ddr, 0x0f, 0),       // DDRCDR
             Clock_divider_gpu         (Divider_gpu, 0x0f, 0),       // GPUCDR
             Clock_divider_hdmi        (Divider_hdmi, 0xff, 0),      // HDMICDR
             Clock_divider_i2s0        (Divider_i2s0, 0xff, 0),      // I2SCDR
             Clock_divider_i2s1        (Divider_i2s1, 0xff, 0),      // I2SCDR
             Clock_divider_lcd0        (Divider_lcd0, 0xff, 0),      // LPCDR
             Clock_divider_lcd1        (Divider_lcd1, 0xff, 0),      // LPCDR
             Clock_divider_msc0        (Divider_msc0, 0xff, 0),      // MSC0CDR
             Clock_divider_msc1        (Divider_msc1, 0xff, 0),      // MSC1CDR
             Clock_divider_msc2        (Divider_msc2, 0xff, 0),      // MSC2CDR
             Clock_divider_pcm         (Divider_pcm, 0xff, 0),       // PCMCDR
             Clock_divider_ssi         (Divider_ssi, 0xff, 0),       // SSICDR
             Clock_divider_uhc         (Divider_uhc, 0xff, 0),       // UHCCDR
             Clock_divider_vpu         (Divider_vpu, 0x0f, 0),       // VPUCDR

             Clock_gate_main           (Clock_control, 1, 23, true), // GATE_SCLKA
             Clock_gate_ddr            (Clock_gate0, 3, 30, true),   // DDR1, DDR0
             Clock_gate_ipu            (Clock_gate0, 1, 29, true),   // IPU
             Clock_gate_lcd            (Clock_gate0, 3, 27, true),   // LCD, TVE
             Clock_gate_cim            (Clock_gate0, 1, 26, true),   // CIM
             Clock_gate_i2c2           (Clock_gate0, 1, 25, true),   // SMB2
             Clock_gate_uhc            (Clock_gate0, 1, 24, true),   // UHC
             Clock_gate_mac            (Clock_gate0, 1, 23, true),   // MAC
             Clock_gate_gps            (Clock_gate0, 1, 22, true),   // GPS
             Clock_gate_dma            (Clock_gate0, 1, 21, true),   // PDMA
             Clock_gate_ssi2           (Clock_gate0, 1, 20, true),   // SSI2
             Clock_gate_ssi1           (Clock_gate0, 1, 19, true),   // SSI1
             Clock_gate_uart3          (Clock_gate0, 1, 18, true),   // UART3
             Clock_gate_uart2          (Clock_gate0, 1, 17, true),   // UART2
             Clock_gate_uart1          (Clock_gate0, 1, 16, true),   // UART1
             Clock_gate_uart0          (Clock_gate0, 1, 15, true),   // UART0
             Clock_gate_sadc           (Clock_gate0, 1, 14, true),   // SADC
             Clock_gate_kbc            (Clock_gate0, 1, 13, true),   // KBC
             Clock_gate_msc2           (Clock_gate0, 1, 12, true),   // MSC2
             Clock_gate_msc1           (Clock_gate0, 1, 11, true),   // MSC1
             Clock_gate_owi            (Clock_gate0, 1, 10, true),   // OWI
             Clock_gate_tssi0          (Clock_gate0, 1, 9, true),    // TSSI0
             Clock_gate_aic0           (Clock_gate0, 1, 8, true),    // AIC0
             Clock_gate_scc            (Clock_gate0, 1, 7, true),    // SCC
             Clock_gate_i2c1           (Clock_gate0, 1, 6, true),    // SMB1
             Clock_gate_i2c0           (Clock_gate0, 1, 5, true),    // SMB0
             Clock_gate_ssi0           (Clock_gate0, 1, 4, true),    // SSI0
             Clock_gate_msc0           (Clock_gate0, 1, 3, true),    // MSC0
             Clock_gate_otg0           (Clock_gate0, 1, 2, true),    // OTG0
             Clock_gate_bch            (Clock_gate0, 1, 1, true),    // BCH
             Clock_gate_nemc           (Clock_gate0, 1, 0, true),    // NEMC
             Clock_gate_p1             (Clock_gate1, 1, 15, true),   // P1
             Clock_gate_x2d            (Clock_gate1, 1, 14, true),   // X2D
             Clock_gate_des            (Clock_gate1, 1, 13, true),   // DES
             Clock_gate_i2c4           (Clock_gate1, 1, 12, true),   // SMB4
             Clock_gate_ahb_mon        (Clock_gate1, 1, 11, true),   // AHB_MON
             Clock_gate_uart4          (Clock_gate1, 1, 10, true),   // UART4
             Clock_gate_hdmi           (Clock_gate1, 1, 9, true),    // HDMI
             Clock_gate_otg1           (Clock_gate1, 1, 8, true),    // OTG1
             Clock_gate_gpvlc          (Clock_gate1, 1, 7, true),    // GPVLC
             Clock_gate_aic1           (Clock_gate1, 1, 6, true),    // AIC1
             Clock_gate_compress       (Clock_gate1, 1, 5, true),    // COMPRESS
             Clock_gate_gpu            (Clock_gate1, 1, 4, true),    // GPU
             Clock_gate_pcm            (Clock_gate1, 1, 3, true),    // PCM
             Clock_gate_vpu            (Clock_gate1, 1, 2, true),    // VPU
             Clock_gate_tssi1          (Clock_gate1, 1, 1, true),    // TSSI1
             Clock_gate_i2c3           (Clock_gate1, 1, 0, true),    // I2C3

             Pll_enable_A              (Pll_control_A, 1, 0),        // APLLEN
             Pll_enable_E              (Pll_control_E, 1, 0),        // EPLLEN
             Pll_enable_M              (Pll_control_M, 1, 0),        // MPLLEN
             Pll_enable_V              (Pll_control_V, 1, 0),        // VPLLEN

             Pll_stable_A              (Pll_control_A, 1, 4),        // APLL_ON
             Pll_stable_E              (Pll_control_E, 1, 4),        // EPLL_ON
             Pll_stable_M              (Pll_control_M, 1, 4),        // MPLL_ON
             Pll_stable_V              (Pll_control_V, 1, 4),        // VPLL_ON

             Pll_bypass_A              (Pll_control_A, 1, 1),        // APLL_BP
             Pll_bypass_E              (Pll_control_E, 1, 1),        // EPLL_BP
             Pll_bypass_M              (Pll_control_M, 1, 1),        // MPLL_BP
             Pll_bypass_V              (Pll_control_V, 1, 1),        // VPLL_BP

             Pll_multiplier_A          (Pll_control_A, 0x1fff, 19),  // APLLM
             Pll_multiplier_E          (Pll_control_E, 0x1fff, 19),  // EPLLM
             Pll_multiplier_M          (Pll_control_M, 0x1fff, 19),  // MPLLM
             Pll_multiplier_V          (Pll_control_V, 0x1fff, 19),  // VPLLM

             Pll_input_division_A      (Pll_control_A, 0x3f, 13),    // APLLN
             Pll_input_division_E      (Pll_control_E, 0x3f, 13),    // EPLLN
             Pll_input_division_M      (Pll_control_M, 0x3f, 13),    // MPLLN
             Pll_input_division_V      (Pll_control_V, 0x3f, 13),    // VPLLN

             Pll_output_division_A     (Pll_control_A, 0x0f, 9),     // APLLOD
             Pll_output_division_E     (Pll_control_E, 0x0f, 9),     // EPLLOD
             Pll_output_division_M     (Pll_control_M, 0x0f, 9),     // MPLLOD
             Pll_output_division_V     (Pll_control_V, 0x0f, 9);     // VPLLOD



// Multiplexer instances.

#define Clocks(...) ((enum Clock_identifiers []) {__VA_ARGS__})

static Mux mux_external (Clock_external),

           // Clocks being propagated to others.

           mux_clock_ssi      (Clock_ssi),
           mux_clock_msc      (Clock_msc),
           mux_hclock2        (Clock_hclock2),
           mux_hclock2_pclock (Clock_hclock2_pclock),
           mux_pclock         (Clock_pclock),

           // Main bus and peripheral clock sources.

           mux_ahb2     (4, Clocks(Clock_none, Clock_main, Clock_pll_M, Clock_rtc)),
           mux_core     (4, Clocks(Clock_none, Clock_main, Clock_pll_M, Clock_pll_E)),
           mux_main     (4, Clocks(Clock_none, Clock_pll_A, Clock_external, Clock_rtc)),

           // Memory and device clock sources.

           mux_cim      (2, Clocks(Clock_main, Clock_pll_M)),
           mux_dev      (3, Clocks(Clock_none, Clock_main, Clock_pll_M)),
           mux_lcd      (3, Clocks(Clock_main, Clock_pll_M, Clock_pll_V)),
           mux_usb      (3, Clocks(Clock_main, Clock_pll_M, Clock_pll_E /* , OTG PHY */)),

           // Clock selectors involving the external clock.

           mux_i2s      (4, Clocks(Clock_external, Clock_external, Clock_main, Clock_pll_E)),
           mux_pcm      (8, Clocks(Clock_external, Clock_external, Clock_external, Clock_external,
                                   Clock_main, Clock_pll_M, Clock_pll_E, Clock_pll_V)),
           mux_ssi      (4, Clocks(Clock_external, Clock_external, Clock_main, Clock_pll_M));



// Clock instances.

static Clock_null        clock_none;

static Clock_passive     clock_external(48000000), clock_rtc(32768);



// Note the use of extra parentheses due to the annoying C++ "most vexing parse"
// problem. See: https://en.wikipedia.org/wiki/Most_vexing_parse

static Clock clock_ahb_mon((Source(mux_external)), (Control(Clock_gate_ahb_mon))),

             clock_compress((Source(mux_external)), (Control(Clock_gate_compress))),

             clock_des((Source(mux_external)), (Control(Clock_gate_des))),

             clock_dma((Source(mux_external)), (Control(Clock_gate_dma))),

             clock_gps((Source(mux_external)), (Control(Clock_gate_gps))),

             clock_gpvlc((Source(mux_external)), (Control(Clock_gate_gpvlc))),

             clock_i2c0((Source(mux_pclock)), (Control(Clock_gate_i2c0))),

             clock_i2c1((Source(mux_pclock)), (Control(Clock_gate_i2c1))),

             clock_i2c2((Source(mux_pclock)), (Control(Clock_gate_i2c2))),

             clock_i2c3((Source(mux_pclock)), (Control(Clock_gate_i2c3))),

             clock_i2c4((Source(mux_pclock)), (Control(Clock_gate_i2c4))),

             clock_i2s0(Source(mux_i2s, Clock_source_i2s0), Control(Clock_gate_aic0)),

             clock_i2s1(Source(mux_i2s, Clock_source_i2s1), Control(Clock_gate_aic1)),

             clock_ipu((Source(mux_external)), (Control(Clock_gate_ipu))),

             clock_kbc((Source(mux_external)), (Control(Clock_gate_kbc))),

             clock_lcd((Source(mux_external)), (Control(Clock_gate_lcd))),

             clock_main(Source(mux_main, Clock_source_main), Control(Clock_gate_main)),

             clock_mac((Source(mux_external)), (Control(Clock_gate_mac))),

             clock_msc((Source(mux_dev, Clock_source_msc))),

             clock_nemc((Source(mux_hclock2)), (Control(Clock_gate_nemc))),

             clock_otg0((Source(mux_external)), (Control(Clock_gate_otg0))),

             clock_otg1((Source(mux_external)), (Control(Clock_gate_otg1))),

             clock_owi((Source(mux_external)), (Control(Clock_gate_owi))),

             clock_sadc((Source(mux_external)), (Control(Clock_gate_sadc))),

             clock_scc((Source(mux_external)), (Control(Clock_gate_scc))),

             clock_tssi0((Source(mux_external)), (Control(Clock_gate_tssi0))),

             clock_tssi1((Source(mux_external)), (Control(Clock_gate_tssi1))),

             clock_uart0((Source(mux_external)), (Control(Clock_gate_uart0))),

             clock_uart1((Source(mux_external)), (Control(Clock_gate_uart1))),

             clock_uart2((Source(mux_external)), (Control(Clock_gate_uart2))),

             clock_uart3((Source(mux_external)), (Control(Clock_gate_uart3))),

             clock_x2d((Source(mux_external)), (Control(Clock_gate_x2d))),

             // Special parent clock for hclock2 and pclock.

             clock_hclock2_pclock(Source(mux_ahb2, Clock_source_hclock2)),

             // SSI channel clocks depending on a common parent divider.

             clock_ssi0((Source(mux_clock_ssi)), Control(Clock_gate_ssi0)),

             clock_ssi1((Source(mux_clock_ssi)), Control(Clock_gate_ssi1)),

             clock_ssi2((Source(mux_clock_ssi)), Control(Clock_gate_ssi2));

static Clock_divided
             clock_bch(Source(mux_core, Clock_source_bch),
                       Control(Clock_gate_bch, Clock_change_enable_bch, Clock_busy_bch),
                       Divider(Clock_divider_bch)),

             clock_cim(Source(mux_cim, Clock_source_cim),
                       Control(Clock_gate_cim, Clock_change_enable_cim, Clock_busy_cim),
                       Divider(Clock_divider_cim)),

             clock_cpu(Source(mux_core, Clock_source_cpu),
                       Control(Field::undefined, Clock_change_enable_cpu, Clock_busy_cpu),
                       Divider(Clock_divider_cpu)),

             clock_ddr(Source(mux_dev, Clock_source_ddr),
                       Control(Clock_gate_ddr, Clock_change_enable_ddr, Clock_busy_ddr),
                       Divider(Clock_divider_ddr)),

             clock_gpu(Source(mux_core, Clock_source_gpu),
                       Control(Clock_gate_gpu, Clock_change_enable_gpu, Clock_busy_gpu),
                       Divider(Clock_divider_gpu)),

             clock_hclock0(Source(mux_core, Clock_source_hclock0),
                           Control(Field::undefined, Clock_change_enable_ahb0),
                           Divider(Clock_divider_hclock0)),

             clock_hclock2(Source(mux_hclock2_pclock),
                           Control(Field::undefined, Clock_change_enable_ahb2),
                           Divider(Clock_divider_hclock2)),

             clock_hdmi(Source(mux_lcd, Clock_source_hdmi),
                        Control(Clock_gate_hdmi, Clock_change_enable_hdmi, Clock_busy_hdmi),
                        Divider(Clock_divider_hdmi)),

             clock_lcd_pixel0(Source(mux_lcd, Clock_source_lcd0),
                              Control(Clock_gate_lcd, Clock_change_enable_lcd0, Clock_busy_lcd0),
                              Divider(Clock_divider_lcd0)),

             clock_lcd_pixel1(Source(mux_lcd, Clock_source_lcd1),
                              Control(Clock_gate_lcd, Clock_change_enable_lcd1, Clock_busy_lcd1),
                              Divider(Clock_divider_lcd1)),

             clock_msc0(Source(mux_clock_msc),
                        Control(Clock_gate_msc0, Clock_change_enable_msc0, Clock_busy_msc0),
                        Divider(Clock_divider_msc0)),

             clock_msc1(Source(mux_clock_msc),
                        Control(Clock_gate_msc1, Clock_change_enable_msc1, Clock_busy_msc1),
                        Divider(Clock_divider_msc1)),

             clock_msc2(Source(mux_clock_msc),
                        Control(Clock_gate_msc2, Clock_change_enable_msc2, Clock_busy_msc2),
                        Divider(Clock_divider_msc2)),

             clock_pcm(Source(mux_pcm, Clock_source_pcm),
                       Control(Clock_gate_pcm, Clock_change_enable_pcm, Clock_busy_pcm),
                       Divider(Clock_divider_pcm)),

             clock_pclock(Source(mux_hclock2_pclock),
                          Control(Field::undefined, Clock_change_enable_ahb2),
                          Divider(Clock_divider_pclock)),

             clock_ssi(Source(mux_ssi, Clock_source_ssi),
                       Control(Field::undefined, Clock_change_enable_ssi, Clock_busy_ssi),
                       Divider(Clock_divider_ssi)),

             clock_uhc(Source(mux_usb, Clock_source_uhc),
                        Control(Clock_gate_uhc, Clock_change_enable_uhc, Clock_busy_uhc),
                        Divider(Clock_divider_uhc));



const double jz4780_pll_intermediate_min = 300000000,
             jz4780_pll_intermediate_max = 1500000000;

static Pll   clock_pll_A(Source(mux_external),
                         Control_pll(Pll_enable_A, Pll_stable_A, Pll_bypass_A),
                         Divider_pll(Pll_multiplier_A, Pll_input_division_A,
                                     Pll_output_division_A,
                                     jz4780_pll_intermediate_min, jz4780_pll_intermediate_max)),

             clock_pll_E(Source(mux_external),
                         Control_pll(Pll_enable_E, Pll_stable_E, Pll_bypass_E),
                         Divider_pll(Pll_multiplier_E, Pll_input_division_E,
                                     Pll_output_division_E,
                                     jz4780_pll_intermediate_min, jz4780_pll_intermediate_max)),

             clock_pll_M(Source(mux_external),
                         Control_pll(Pll_enable_M, Pll_stable_M, Pll_bypass_M),
                         Divider_pll(Pll_multiplier_M, Pll_input_division_M,
                                     Pll_output_division_M,
                                     jz4780_pll_intermediate_min, jz4780_pll_intermediate_max)),

             clock_pll_V(Source(mux_external),
                         Control_pll(Pll_enable_V, Pll_stable_V, Pll_bypass_V),
                         Divider_pll(Pll_multiplier_V, Pll_input_division_V,
                                     Pll_output_division_V,
                                     jz4780_pll_intermediate_min, jz4780_pll_intermediate_max));



// Clock register.

static Clock_base *clocks[Clock_identifier_count] = {
  &clock_none,            // Clock_aic
  &clock_none,            // Clock_aic_bitclk
  &clock_none,            // Clock_aic_pclk
  &clock_none,            // Clock_can0
  &clock_none,            // Clock_can1
  &clock_none,            // Clock_cdbus
  &clock_cim,
  &clock_cpu,
  &clock_ddr,
  &clock_dma,
  &clock_none,            // Clock_emac
  &clock_external,
  &clock_hclock0,
  &clock_hclock2,
  &clock_hclock2_pclock,
  &clock_hdmi,
  &clock_i2c0,
  &clock_i2c1,
  &clock_i2c2,
  &clock_i2c3,
  &clock_i2c4,
  &clock_i2s0,
  &clock_none,            // Clock_i2s0_rx
  &clock_none,            // Clock_i2s0_tx
  &clock_i2s1,
  &clock_none,            // Clock_i2s1_rx
  &clock_none,            // Clock_i2s1_tx
  &clock_none,            // Clock_kbc
  &clock_lcd,
  &clock_lcd_pixel0,
  &clock_lcd_pixel1,
  &clock_mac,
  &clock_main,
  &clock_none,            // Clock_mipi_csi
  &clock_msc,
  &clock_msc0,
  &clock_msc1,
  &clock_msc2,
  &clock_none,            // Clock_none
  &clock_otg0,
  &clock_otg1,
  &clock_pclock,
  &clock_pll_A,
  &clock_pll_E,
  &clock_pll_M,
  &clock_pll_V,
  &clock_none,            // Clock_pwm0
  &clock_none,            // Clock_pwm1
  &clock_rtc,
  &clock_scc,
  &clock_none,            // Clock_sfc
  &clock_ssi,
  &clock_ssi0,
  &clock_ssi1,
  &clock_ssi2,
  &clock_none,            // Clock_timer
  &clock_uart0,
  &clock_uart1,
  &clock_uart2,
  &clock_uart3,
  &clock_none,            // Clock_udc
  &clock_uhc,
  &clock_none,            // Clock_uprt
};



// Peripheral abstraction.

Cpm_jz4780_chip::Cpm_jz4780_chip(l4_addr_t addr)
: Cpm_chip(addr, clocks)
{
}



// C language interface functions.

void
*jz4780_cpm_init(l4_addr_t cpm_base)
{
  return (void *) new Cpm_jz4780_chip(cpm_base);
}

int
jz4780_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->have_clock(clock);
}

void
jz4780_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->start_clock(clock);
}

void
jz4780_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->stop_clock(clock);
}

int
jz4780_cpm_get_parameters(void *cpm, enum Clock_identifiers clock, uint32_t parameters[])
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_parameters(clock, parameters);
}

int
jz4780_cpm_set_parameters(void *cpm, enum Clock_identifiers clock, int num_parameters, uint32_t parameters[])
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->set_parameters(clock, num_parameters, parameters);
}

uint8_t
jz4780_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_source(clock);
}

void
jz4780_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{ 
  static_cast<Cpm_jz4780_chip *>(cpm)->set_source(clock, source);
}

enum Clock_identifiers
jz4780_cpm_get_source_clock(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_source_clock(clock);
}

void
jz4780_cpm_set_source_clock(void *cpm, enum Clock_identifiers clock, enum Clock_identifiers source)
{ 
  static_cast<Cpm_jz4780_chip *>(cpm)->set_source_clock(clock, source);
}

uint32_t
jz4780_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_source_frequency(clock);
}

uint32_t
jz4780_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_frequency(clock);
}

int
jz4780_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->set_frequency(clock, frequency);
}
